"""

``evaluate()`` is used to make an evaluation of a trained instance of
``globalemu``. It has to be initialised with a set of kwargs, most importantly
the ``base_dir`` which contains the trained model. Once initialised it can
then be used to make predictions and return the predicted signal plus the
corresponding redshift. ``evaluate()`` can reproduce a high resolution Global
21-cm signal (450 redshift data points) in 1.5 ms.

"""

import tensorflow as tf
import numpy as np
from tensorflow import keras
from tensorflow.keras import backend as K
import gc
import pickle


class evaluate():

    r"""

    The class can be initialised with the following kwargs and the
    following code

    .. code:: python

        predictor = evaluate(**kwargs)

    **kwargs:**

        base_dir: **string / default: 'model_dir/'**
            | The ``base_dir`` is where the trained model is saved.

        model: **tensorflow model / default: None**
            | If making multiple calls to the function it is advisable to
                load the trained model in the script making the calls and
                then to pass it to ``evaluate()``. This prevents the model
                being loaded upon each call and leads to a significant
                increase speed. You can load a model via,

                .. code:: python

                    from tensorflow import keras

                    model = keras.models.load_model(
                        base_dir + 'model.h5',
                        compile=False)

        logs: **list / default: [0, 1, 2]**
            | The indices corresponding to the astrophysical
                parameters that
                were logged during training. The default assumes
                that the first three columns in "train_data.txt" are
                :math:`{f_*}` (star formation efficiency),
                :math:`{V_c}` (minimum virial circular velocity) and
                :math:`{f_x}` (X-ray efficieny).

        gc: **Bool / default: False**
            | Multiple calls to the function can cause runaway memory
                related issues (it is worth testing this
                behaviour before scheduling
                hpc jobs) and these memory issues can be somewhat eleviated
                by setting ``gc=True``. This performs a garbage collection
                after every function call. It is an optional argumen set to
                ``False`` by default because it can increase the time taken
                to perform the emulation.

        z: **list or np.array / default: Original redshift array**
            | The redshift values at which you want to emulate the 21-cm
                signal. The default is given by the redshift range that the
                network was originally trained on (found in ``base_dir``).

    Once the class has been initialised you can then make evaluations
    of the emulator by passing the parameters like so

    .. code:: python

        signal, z = predictor(parameters)

    **Parameters:**

        parameters: **list or np.array**
            | The combination of astrophysical parameters that you want to
                emulate a global signal for. They must be in the same order
                as was used when training and they must fall with in the
                trained parameter space. For the 21cmGEM data the order
                of the astrophysical parameters is given by:
                :math:`{f_*, V_c, f_x, \tau, \alpha, \nu_\mathrm{min}}` and
                :math:`{R_\mathrm{mfp}}` (see the ``globalemu`` paper and
                references therein for a description of the parameters).
                You can pass a single set of parameters or a 2D array of
                different parameters to evaluate. For example if I wanted to
                evaluate 100 sets of 7 parameters my input array should have
                shape=(100, 7).

    **Return:**

        signal: **array or float**
            | The emulated signal. If a single redshift is passed to the
                emulator then the returned signal will be a single float
                otherwise the result will be an array. If more than
                one set of parameters are input then the output signal will
                be an array of signals. e.g. 100 input sets of parameters
                gives signal.shape=(100, len(z)).

        z: **array or float**
            | The redshift values corresponding to the returned signal. If
                z was not specified on input then the returned signal and
                redshifts will correspond to the redshifts that the network
                was originally trained on.

    """

    def __init__(self, **kwargs):

        for key, values in kwargs.items():
            if key not in set(
                    ['base_dir', 'model', 'logs', 'gc', 'z']):
                raise KeyError("Unexpected keyword argument in evaluate()")

        self.base_dir = kwargs.pop('base_dir', 'model_dir/')
        if type(self.base_dir) is not str:
            raise TypeError("'base_dir' must be a sting.")
        elif self.base_dir.endswith('/') is False:
            raise KeyError("'base_dir' must end with '/'.")

        file = open(self.base_dir + "preprocess_settings.pkl", "rb")
        self.preprocess_settings = pickle.load(file)

        self.data_mins = np.loadtxt(self.base_dir + 'data_mins.txt')
        self.data_maxs = np.loadtxt(self.base_dir + 'data_maxs.txt')
        if self.preprocess_settings['resampling'] is True:
            self.cdf = np.loadtxt(self.base_dir + 'cdf.txt')

        self.model = kwargs.pop('model', None)
        if self.model is None:
            self.model = keras.models.load_model(
                self.base_dir + 'model.h5',
                compile=False)

        self.logs = kwargs.pop('logs', [0, 1, 2])
        if type(self.logs) is not list:
            raise TypeError("'logs' must be a list.")
        self.garbage_collection = kwargs.pop('gc', False)

        boolean_kwargs = [self.garbage_collection]
        boolean_strings = ['gc']
        for i in range(len(boolean_kwargs)):
            if type(boolean_kwargs[i]) is not bool:
                raise TypeError("'" + boolean_strings[i] + "' must be a bool.")

        if self.preprocess_settings['AFB'] is True:
            self.AFB = np.loadtxt(self.base_dir + 'AFB.txt')
        if self.preprocess_settings['std_division'] is True:
            self.label_stds = np.load(self.base_dir + 'labels_stds.npy')

        self.original_z = np.loadtxt(self.base_dir + 'z.txt')

        self.z = kwargs.pop('z', self.original_z)
        if type(self.z) not in set([np.ndarray, list, int, float]):
            raise TypeError("'z' should be a numpy array, list, float or int.")

    def __call__(self, parameters):

        if type(parameters) not in set([np.ndarray, list]):
            raise TypeError("'params' must be a list or np.array.")
        if type(parameters) is list:
            parameters = np.array(parameters)

        if len(parameters.shape) == 1:
            params = []
            for i in range(len(parameters)):
                if i in set(self.logs):
                    if parameters[i] == 0:
                        parameters[i] = 1e-6
                    params.append(np.log10(parameters[i]))
                else:
                    params.append(parameters[i])
            normalised_params = np.array([
                (params[i] - self.data_mins[i]) /
                (self.data_maxs[i] - self.data_mins[i])
                for i in range(len(params))])
        else:
            params = []
            for i in range(len(parameters)):
                params_set = []
                for j in range(len(parameters[i])):
                    if j in set(self.logs):
                        if parameters[i, j] == 0:
                            parameters[i, j] = 1e-6
                        params_set.append(np.log10(parameters[i, j]))
                    else:
                        params_set.append(parameters[i, j])
                params.append(params_set)
            params = np.array(params)
            normalised_params = np.array([
                (params[:, i] - self.data_mins[i]) /
                (self.data_maxs[i] - self.data_mins[i])
                for i in range(params.shape[1])]).T

        if self.preprocess_settings['resampling'] is True:
            norm_z = np.interp(self.z, self.original_z, self.cdf)
        else:
            norm_z = (self.z - self.original_z.min()) / \
                     (self.original_z.max() - self.original_z.min())
        if isinstance(norm_z, np.ndarray):
            if len(normalised_params.shape) == 1:
                x = np.tile(normalised_params, (len(norm_z), 1))
                x = np.hstack([x, norm_z[:, np.newaxis]])
            else:
                x = np.hstack([
                               np.vstack([np.tile(normalised_params[i],
                                         (len(norm_z), 1))
                                         for i in
                                         range(len(normalised_params))]),
                               np.vstack([norm_z[:, np.newaxis]] *
                                         normalised_params.shape[0])])
            tensor = tf.convert_to_tensor(x, dtype=tf.float32)
            result = self.model(tensor, training=False).numpy()
            if len(normalised_params.shape) != 1:
                evaluation = np.array([
                    result[i:i + len(self.z)]
                    for i in range(0, len(result), len(self.z))])[:, :, 0]
            else:
                evaluation = result.T[0]
            if self.garbage_collection is True:
                K.clear_session()
                gc.collect()
        else:
            x = np.hstack([normalised_params, norm_z]).astype(np.float32)
            result = self.model(x[np.newaxis, :], training=False).numpy()
            evaluation = result[0][0]

        if self.preprocess_settings['std_division'] is True:
            if isinstance(evaluation, np.ndarray):
                evaluation = [
                    evaluation[i]*self.label_stds
                    for i in range(evaluation.shape[0])]
            else:
                evaluation *= self.label_stds

        if self.preprocess_settings['AFB'] is True:
            evaluation += np.interp(self.z, self.original_z, self.AFB)

        if type(evaluation) is not np.ndarray:
            evaluation = np.array(evaluation)
        
        return evaluation, self.z



from tensorflow.keras import backend as K


class loss_functions():

    r"""

    The loss functions used to assess the quality of the emulation.

    **Parameters:**

    y: **tf.tensor**
        | The training labels for a given batch of training data.

    y_: **tf.tensor**
        | The predicted value of the training labels corresponding to y.

    """

    def __init__(self, y, y_):
        self.y = y
        self.y_ = y_

    def rmse(self):
        return K.sqrt(K.mean(K.square(self.y - self.y_)))

    def mse(self):
        return K.mean(K.square(self.y - self.y_))

    def GEMLoss(self):
        """
        This loss function corresponds to eq. 11 in the 21cmGEM paper
        (https://arxiv.org/abs/1910.06274) and
        is used in the globalemu MNRAS paper to compare the performance of the
        two emulators.
        """
        return (
            K.sqrt(K.mean(K.square(self.y - self.y_))) /
            K.max(K.abs(self.y)))



import tensorflow as tf


class network_models():

    r"""

    This function is used during training to generate the tensorflow model
    for training.

    """
    def __init__(self):
        self.Model = tf.keras.models.Model
        self.Inputs = tf.keras.layers.Input
        self.Dense = tf.keras.layers.Dense
        self.Dropout = tf.keras.layers.Dropout

    def basic_model(
            self, input_dim, output_dim, layer_sizes, activation,
            drop_val, output_activation):

        r"""

        The basic tensorflow model used in globalemu. The parameters are set
        in network.py.

        **Parameters:**

        input_dim: **int**
            | Number of input nodes for the network.

        output_dim: **int**
            | Number of output nodes for the network.

        layer_size: **list**
            | A list containing the structure of the hidden layers in the
                network. The number of elements in the list corresponds
                to the number of hidden layers and the value of each element
                to the number of nodes in that layer. e.g. [12, 12]
                corresponds to 2 hidden layers each with 12 nodes.

        """
        a0 = self.Inputs(shape=(input_dim,))
        inputs = a0
        for layer_size in layer_sizes:
            outputs = self.Dense(layer_size, activation=activation)(a0)
            outputs = self.Dropout(drop_val)(outputs)
            a0 = outputs
        outputs = self.Dense(output_dim, activation=output_activation)(a0)
        model = self.Model(inputs, outputs)
        return model



"""

``process()`` is used to preprocess the data in the provided directory
using the techniques outlined in the ``globalemu`` paper. For ``process()``
to work it requires the testing and training data to be saved in the
``data_location`` directory in a specific manner. The "labels" or temperatures
(network outputs) should be saved
as "test_labels.txt"/"train_labels.txt" and the "data" or
astrophysical parameters (network inputs excluding redshift) as
"test_data.txt"/"train_data.txt".

"""

import numpy as np
import os
import pandas as pd
import pickle
from globalemu.cmSim import calc_signal
from globalemu.resample import sampling


class process():

    r"""

    **Parameters:**

        num: **int**
            | The number of models that will be used to train globalemu. If
                you wish to use the full training data set then set
                ``num = 'full'``.

        z: **np.array**
            | The redshift range that corresponds to the models in the saved
                "test_labels.txt" and "train_labels.txt" e.g. for the
                ``21cmGEM`` data this would be :code:`np.arange(5, 50.1, 0.1)`.

    **kwargs:**

        base_dir: **string / default: 'model_dir/'**
            | The ``base_dir`` is where the preprocessed data and later the
                trained models will be placed. This should be thought of as the
                working directory as it will be needed when training a model
                and making evaluations of trained models.

        data_location: **string / default: 'data/'**
            | As discussed above the ``data_loaction`` is where the data to be
                processed is to be found. It must be accurately provided for
                the code to work and must end in a '/'.

        xHI: **Bool / default: False**
            | If True then ``globalemu`` will act as if it is training a
                neutral fraction history emulator.

        AFB: **Bool / default: None**
            | If True then ``globalemu`` will calculate an astrophysics free
                baseline and subtract this from the training data signals.
                The AFB is specific to the global 21-cm signal and as
                ``globalemu`` is set up to emulate the global signal by
                default this parameter is set to True. If xHI is True then
                AFB is set to False by default.

        std_division: **Bool / default: None**
            | If True then ``globalemu`` will divide the training data by the
                standard deviation across the training data. This is
                recommended when building an emulator to emulate the global
                signal and is set to True by default. If xHI is True then
                std_division is set to False by default.

        resampling: **Bool / default: None**
            | Controls whether or not the signals will be resampled with
                higher sampling at regions of large variation in the training
                data set or not. Set to True by default as this is advised for
                training both neutral fraction and global signal emulators.

        logs: **list / default: [0, 1, 2]**
            | The indices corresponding to the astrophysical parameters in
                "train_data.txt" that need to be logged. The default assumes
                that the first three columns in "train_data.txt" are
                :math:`{f_*}` (star formation efficiency),
                :math:`{V_c}` (minimum virial circular velocity) and
                :math:`{f_x}` (X-ray efficieny).
    """

    def __init__(self, num, z, **kwargs):

        print('Preprocessing started...')

        for key, values in kwargs.items():
            if key not in set(
                    ['base_dir', 'data_location', 'xHI', 'logs', 'AFB',
                     'std_division', 'resampling']):
                raise KeyError("Unexpected keyword argument in process()")

        self.num = num
        if type(self.num) is not int:
            if self.num != 'full':
                raise TypeError("'num' must be an integer or 'full'.")

        self.z = z
        if type(self.z) not in set([np.ndarray, list]):
            raise TypeError("'z' should be a numpy array or list.")

        self.base_dir = kwargs.pop('base_dir', 'model_dir/')
        self.data_location = kwargs.pop('data_location', 'data/')

        file_kwargs = [self.base_dir, self.data_location]
        file_strings = ['base_dir', 'data_location']
        for i in range(len(file_kwargs)):
            if type(file_kwargs[i]) is not str:
                raise TypeError("'" + file_strings[i] + "' must be a sting.")
            elif file_kwargs[i].endswith('/') is False:
                raise KeyError("'" + file_strings[i] + "' must end with '/'.")

        self.xHI = kwargs.pop('xHI', False)
        if self.xHI is False:
            self.preprocess_settings = {'AFB': True, 'std_division': True,
                                        'resampling': True}
        else:
            self.preprocess_settings = {'AFB': False, 'std_division': False,
                                        'resampling': True}

        preprocess_settings_keys = ['AFB', 'std_division', 'resampling']
        for key in preprocess_settings_keys:
            if key in kwargs:
                self.preprocess_settings[key] = kwargs[key]

        bool_kwargs = [self.xHI, self.preprocess_settings['AFB'],
                       self.preprocess_settings['std_division'],
                       self.preprocess_settings['resampling']]
        bool_strings = ['xHI', 'AFB', 'std_division', 'resampling']
        for i in range(len(bool_kwargs)):
            if type(bool_kwargs[i]) is not bool:
                raise TypeError(bool_strings[i] + " must be a bool.")

        self.logs = kwargs.pop('logs', [0, 1, 2])
        if type(self.logs) is not list:
            raise TypeError("'logs' must be a list.")

        if not os.path.exists(self.base_dir):
            os.mkdir(self.base_dir)

        file = open(self.base_dir + "preprocess_settings.pkl", "wb")
        pickle.dump(self.preprocess_settings, file)
        file.close()

        np.savetxt(self.base_dir + 'z.txt', self.z)

        full_train_data = pd.read_csv(
            self.data_location + 'train_data.txt',
            delim_whitespace=True, header=None).values
        full_train_labels = pd.read_csv(
            self.data_location + 'train_labels.txt',
            delim_whitespace=True, header=None).values

        if self.preprocess_settings['AFB'] is True:
            np.save(
                self.base_dir + 'AFB_norm_factor.npy',
                full_train_labels[0, -1]*1e-3)
            res = calc_signal(self.z, self.base_dir)

        if self.num == 'full':
            train_data = full_train_data.copy()
            if self.preprocess_settings['AFB'] is True:
                train_labels = full_train_labels.copy() - res.deltaT
            else:
                train_labels = full_train_labels.copy()
        else:
            ind = []
            for i in range(len(full_train_labels)):
                index = np.random.randint(0, len(full_train_labels))
                if index not in set(ind):
                    ind.append(index)
                if len(ind) == self.num:
                    break
            ind = np.array(ind)

            train_data, train_labels = [], []
            for i in range(len(full_train_labels)):
                if np.any(ind == i):
                    train_data.append(full_train_data[i, :])
                    if self.preprocess_settings['AFB'] is True:
                        train_labels.append(full_train_labels[i] - res.deltaT)
                    else:
                        train_labels.append(full_train_labels[i])
            train_data, train_labels = np.array(train_data), \
                np.array(train_labels)

        log_td = []
        for i in range(train_data.shape[1]):
            if i in set(self.logs):
                for j in range(train_data.shape[0]):
                    if train_data[j, i] == 0:
                        train_data[j, i] = 1e-6
                log_td.append(np.log10(train_data[:, i]))
            else:
                log_td.append(train_data[:, i])
        train_data = np.array(log_td).T

        if self.preprocess_settings['resampling'] is True:
            sampling_call = sampling(
                self.z, self.base_dir, train_labels)
            samples = sampling_call.samples
            cdf = sampling_call.cdf

            resampled_labels = []
            for i in range(len(train_labels)):
                resampled_labels.append(
                    np.interp(samples, self.z, train_labels[i]))
            train_labels = np.array(resampled_labels)

            norm_s = np.interp(samples, self.z, cdf)
        else:
            norm_s = (self.z - self.z.min())/(self.z.max() - self.z.min())

        data_mins = train_data.min(axis=0)
        data_maxs = train_data.max(axis=0)

        norm_train_data = []
        for i in range(train_data.shape[1]):
            norm_train_data.append(
                (train_data[:, i] - data_mins[i])/(data_maxs[i]-data_mins[i]))
        norm_train_data = np.array(norm_train_data).T

        if self.preprocess_settings['std_division'] is True:
            labels_stds = train_labels.std()
            norm_train_labels = [
                train_labels[i, :]/labels_stds
                for i in range(train_labels.shape[0])]
            norm_train_labels = np.array(norm_train_labels)

            norm_train_labels = norm_train_labels.flatten()
            np.save(self.base_dir + 'labels_stds.npy', labels_stds)
        else:
            norm_train_labels = train_labels.flatten()

        if self.num != 'full':
            np.savetxt(self.base_dir + 'indices.txt', ind)
        np.savetxt(self.base_dir + 'data_mins.txt', data_mins)
        np.savetxt(self.base_dir + 'data_maxs.txt', data_maxs)

        flattened_train_data = []
        for i in range(len(norm_train_data)):
            for j in range(len(norm_s)):
                flattened_train_data.append(
                    np.hstack([norm_train_data[i, :], norm_s[j]]))
        flattened_train_data = np.array(flattened_train_data)

        train_data, train_label = flattened_train_data, norm_train_labels
        train_dataset = np.hstack([train_data, train_label[:, np.newaxis]])

        np.savetxt(
            self.base_dir + 'train_dataset.csv', train_dataset, delimiter=',')
        np.savetxt(self.base_dir + 'train_data.txt', train_data)
        np.savetxt(self.base_dir + 'train_label.txt', train_label)

        print('...preprocessing done.')
